/*****************************************************************************
 * Orangutan LV-168 Demo program 4:                                          *
 * Temperature sensor, UART communication, and interrupts.                   *
 *****************************************************************************/

// To make full use of this demo, you will need a source capable of producing
// logic-level (TTL) serial transmissions at 115.2 kbps.  The Pololu Orangutan
// USB programmer can be such a source if you switch the blue shorting block to
// the USB-to-serial-adapter position and connect its TX pin to pin PD0 on your
// Orangutan LV-168.  Don't forget to also connect the programmer's ground pad
// (labeled G) to ground on your Orangutan.  Then just start up a terminal
// program such as hyperterm or Tera Term Pro, connect to the COM port of
// your USB-to-serial adapter, set the baud for 115.2 kbps, and start typing
// characters.  You should see the characters appear on the top row of your
// Orangutan's LCD while the bottom row continues to display the board's
// temperature (note that the temperature is just a raw A-to-D conversion and
// is not in any units--you will need to empirically determine how to convert
// this temperature to units like Farenheit or Celcius).

#include <avr/io.h>
#include <avr/interrupt.h>	// we need this if we want to use interrupts
#include "device.h"			// pinout definitions for the Orangutan LV-168
#include <util/delay.h>		// F_CPU is defined in "device.h" above

#include "lcd.h"			// provides routines for using the LCD


// ***** GLOBAL VARIABLES *****
// The following globals are declared as volatile to force our program to
// reload their values from RAM every time they're used.  We do this because
// we change their values from within an interrupt, and if they are not
// volatile our main code might load local copies into registers and never
// notice that our interrupt changed their values in RAM.  Volatile globals
// slow down your program because they require reads from and writes
// to RAM whenever that varible is accessed, but any variable whose value is
// changed by an interrupt should usually be declared as volatile.
volatile unsigned char RXChar;		// char most recently received from UART
volatile unsigned char newRX = 0;	// flag that a new char has been received



// delay for time_ms milliseconds by looping
void delay_ms( unsigned int time_ms )
{
	unsigned int i;

	for ( i = 0; i < time_ms; i++ )
		_delay_ms( 1 );		// _delay_ms() comes from <util/delay.h>
							//  and can only delay for a max of around 13 ms
							//  when the IO clock is 20 MHz.
}



// This is our receive-complete interrupt routine.  Whenever data comes in
// over the UART, normal execution of our program will be halted and we will
// automatically come here.  Our general strategy is to deal with the interrupt
// as quickly as possible and get out so normal execution can resume.  We could
// do a lot more from within the interrupt itself, such as writing the
// character to the LCD, but that would go against the whole "get out of an
// interrupt as quickly as possible" philosophy.  Instead, we set some
// variables to capture the essence of the interrupt event and process those
// variables outside the interrupt in our main program.
ISR( USART_RX_vect )
{
	RXChar = UDR0;			// store the received char
	newRX = 1;				// flag that we have received a new char
}



int main()
{
	lcd_init();	// this function must be called before any other LCD command

	lcd_string("OrgLV168");	// display the string on the LCD
	lcd_gotoxy(0, 1);		// goto the start of LCD line 2
	lcd_string(" Demo 4");

	delay_ms(1500);			// loop-delay for 1.5 seconds
	lcd_clear();


	// ***** UART INITIALIZATION *****

	// Initialize the UART for 115.2k baud, 8-bit characters, no parity,
	// and 1 stop bit.  Enable the UART's receiver and the receive-complete
	// interrupt.

	UCSR0B = 0x90;		// bit 7 set: RX complete interrupt *ENABLED*
						// bit 6 clear: TX complete interrupt disabled
						// bit 5 clear: data register empty interrupt disabled
						// bit 4 set: receiver *ENABLED* (pin PD0 overridden)
						// bit 3 clear: transmitter disabled
						// bit 2 clear: no 9-bit characters
						// bits 0 & 1: only used for 9-bit characters

	UCSR0C = 0x06;		// bits 6 & 7 clear: asynchronous UART
						// bits 4 & 5 clear: parity generation/check disabled
						// bits 3 clear: one stop bit
						// bits 1 & 2 set: 8-bit characters
						// bit 0: not used in asynchronous mode

	UBRR0 = 10;			// baud = IO clock / (16 * (UBRR0 + 1)
						//      = 20 MHz / (16 * 11) = 113.6 kbps (~115.2 kbps)


	// ***** ADC INITIALIZATION *****

	// The temperature sensor can be optionally jumpered to ADC6 (a dedicated
	// analog input on the mega168 referred to as ADC channel 6) using a blob
	// of solder across the SMT jumper pads on the back of the main PCB.
	// The Orangutan LV-168 ships with this solder jumper in place.
	// This program assumes that solder-short connecting ADC6 to the
	// temperature sensor has not been removed.

	ADCSRA = 0x87;		// bit 7 set: ADC enabled
						// bit 6 clear: don't start conversion
						// bit 5 clear: disable autotrigger
						// bit 4: ADC interrupt flag
						// bit 3 clear: disable ADC interrupt
						// bits 0-2 set: ADC clock prescaler is 128

	ADMUX = 0x06;		// bit 7 and 6 clear: voltage ref is Vref pin
						// bit 5 clear: right-adjust result (10-bit ADC)
						// bit 4 not implemented
						// bits 0-3: ADC channel (channel 6)


	sei();				// global interrupt enable (no interrupts are generated
						// until the global interrupt enable is set)
						// interrupts can be globally disabled by calling cli()

	unsigned char c = 0;

	while (1)			// loop forever
	{
		long sum = 0;
		unsigned int avg, i;

		// Here we accumulate 512 conversion results to get an average ADC.
		// According to the mega168 datasheet, it takes approximately 13
		// ADC clock cycles to perform a conversion.  We have configured
		// the ADC run at IO clock / 128 = 20 MHz / 128 = 156 kHz, which
		// means it has a conversion rate of around 10 kHz.  As a result,
		// it should take around 50 ms to accumulate 500 ADC samples.
		for (i = 0; i < 512; i++)
		{
			ADCSRA |= ( 1 << ADSC );			// start conversion
			while ( ADCSRA & ( 1 << ADSC ))		// wait while converting
				;
			sum += ADC;							// add in conversion result
		}
		avg = sum >> 9;		// a.k.a. avg = sum / 2^9 = avg / 512
							//  typically, bitshifting is faster than division

		while (newRX)					// loop until newRX flag stays cleared
		{
			lcd_gotoxy(c, 0);			// go to char 4 of LCD line 1
			c = (c + 1) & 7;			// i.e. c = c + 1; if (c > 7) c = 0;
			newRX = 0;					// clear newRX flag
			lcd_data(RXChar);			// send a single character to the LCD
			if (c == 1)
				lcd_string("       ");	// essentially clear LCD line 1
		}

		lcd_gotoxy(0, 1);				// go to the start of LCD line 2
		lcd_string("Tmp ");
		lcd_int(avg);					// display average ADC as an integer
		lcd_string("   ");				// overwrite any LCD character remnants*/
	}

	return 0;
}
